from upyrc import upyre

# Optional, to have more verbose log
import logging
upyre.set_log_level(logging.DEBUG)

PROJECT_FOLDER = "D:/Projects/UpyreTest/UnrealRemoteControlTestData"
# ==> Change it for your test project path.

# Set your project path here, it will fetch the python settings of your project, but your MUST have the "Enable remote execution" of the python plugin enabled in unreal.
config = upyre.RemoteExecutionConfig.from_uproject_path(f"{PROJECT_FOLDER}/UpyreTest.uproject")

# Create the connection
with upyre.PythonRemoteConnection(config) as conn:
    
    # Example with a file path
    conn.import_fbx_static_meshes(["T:/H_to_EU_FBX/pigtest01.fbx"], "/Game/FBXImportTest")


'''
    Module added to the remote python session (unreal) sys.paths using add_command_json_output_pipe.jinja template if
    upyre.PythonRemoteConnection "open_output_pipe" is set to True or PythonRemoteConnection.init_json_output_pipe() is called.

    This module allows to write out data to a temporary json file in order to be able to read them back from
    the caller session.

    The environment variable "UPYRE_JSON_PIPE_FILE" is used to set up the json file path where the data will be written to.
    It is set or updated using the add_command_json_output_pipe.jinja template as well.

    Once that module is available, the remote python session (unreal) can simple write data using:
    
    from upyre_json_pipe import command_output_pipe
    command_output_pipe.add("EntryName", data)
'''

import json
import os
import tempfile

output_pipe_temp_file = os.environ.get("UPYRE_JSON_PIPE_FILE", tempfile.gettempdir() + "/uepyrc_output_pipe.json")

class CommandOutputPipe():

    def __init__(self, output_pipe_temp_file: str):

        try:
            with open(output_pipe_temp_file, 'r') as f:
                self.__data = json.load(f)
        except:
            self.__data = {}
        
        self._output_pipe_temp_file = output_pipe_temp_file
        dirname = os.path.dirname(str(self._output_pipe_temp_file))
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    def update_temp_file_path(self, output_pipe_temp_file: str):

        self._output_pipe_temp_file = output_pipe_temp_file

    def flush(self):

        self.__data = {}
        try:
            with open(self._output_pipe_temp_file, 'w') as f:
                json.dump(self.__data, f)
        except:
            pass

    def write(self, entry_name, entry_value):
        
        self.__data[entry_name] = entry_value
        try:
            with open(self._output_pipe_temp_file, 'w') as f:
                json.dump(self.__data, f)
        except TypeError:
            str_entry = str(entry_value)
            self.__data[entry_name] = str_entry
            with open(self._output_pipe_temp_file, 'w') as f:
                json.dump(self.__data, f)

    def read(self, entry_name, default=None):
        
        if os.path.exists(self._output_pipe_temp_file):
            with open(self._output_pipe_temp_file, 'r') as f:
                data = json.load(f)
            return data.get(entry_name, default)
        return default

json_pipe = CommandOutputPipe(output_pipe_temp_file)



from upyrc import upyre

# Optional, to have more verbose log
import logging
upyre.set_log_level(logging.DEBUG)

PROJECT_FOLDER = "D:/Projects/UpyreTest/UnrealRemoteControlTestData"
# ==> Change it for your test project path.

# Set your project path here, it will fetch the python settings of your project, but your MUST have the "Enable remote execution" of the python plugin enabled in unreal.
config = upyre.RemoteExecutionConfig.from_uproject_path(f"{PROJECT_FOLDER}/UpyreTest.uproject")

# Create the connection
with upyre.PythonRemoteConnection(config, open_json_output_pipe=True) as conn:
    
    # This execute simple multiline python statement OR file
    result = conn.execute_python_command("unreal.log('Hello')\nunreal.log('World ?')", exec_type=upyre.ExecTypes.EXECUTE_FILE, raise_exc=True)
    print(result)

    # Example with a file path
    conn.write("TestRead", 1234)  # Data sent to process using the json temp file pipe.
    result_file = conn.execute_python_command(f"{PROJECT_FOLDER}/Content/Blueprints/test_pyfile_ue.py", exec_type=upyre.ExecTypes.EXECUTE_FILE)
    print(result_file)

    # Example with a file path and arguments
    result_file = conn.execute_python_command(f"{PROJECT_FOLDER}/Content/Blueprints/test_pyfile_ue.py hello world", exec_type=upyre.ExecTypes.EXECUTE_FILE)
    print(result_file)
    # Here the two arguments "hello" and "world" will be sent to the file before execution. And can be accessed with sys.argv list (the first entry being the file path).
    # >>> From cmd arguments: hello world.

    # You can also evalurate a single line statement and get the result back
    result_statement = conn.execute_python_command("1 + 2", exec_type=upyre.ExecTypes.EVALUATE_STATEMENT)
    print(result_statement) # This print 3

    # Set a property of a utility BP.
    conn.set_bp_property("/Game/Blueprints/BPU_TestUtils.BPU_TestUtils_C", properties={"VarToChange":True})

    # Run utility BP functions.
    conn.execute_bp_method("/Game/Blueprints/BPU_TestUtils.BPU_TestUtils_C", "SimplePrint")
    conn.execute_bp_method("/Game/Blueprints/BPU_TestUtils.BPU_TestUtils_C", "PrintWithArg", args=("Hello", 5), raise_exc=True)

    # Spawn a utility widget BP as tab.
    r = conn.spawn_utility_widget_bp("/Game/Blueprints/BPW_TestUtilWidget.BPW_TestUtilWidget")
    # ---> In Unreal the BPW_TestUtilWidget will be spawned in a tab.
    
    # >>> BPW_TestUtilWidget_C /Game/Levels/L_upyrc.L_upyrc:BPW_TestUtilWidget_C_0
    conn.execute_editor_widget_bp_method("/Game/Blueprints/BPW_TestUtilWidget.BPW_TestUtilWidget", "TestFunction", args=("Hello world",))
    # ---> In Unreal, will call method "TestFunction" of the widget spawned, it will print out "Message: Hello world".
    widget_id = conn.read("WidgetID")
    conn.close_utility_widget_bp_from_id(widget_id)

    # Print out some logs.
    conn.log("I'm a message.")
    conn.log_warning("Uho, I'm a warning...")
    conn.log_error("Oh bummer I'm an error !!!")
    # ---> In Unreal: Will print the logs in the "Output log" tab.

    # Execute custom template.
    conn.execute_template_file(f"{PROJECT_FOLDER}/Content/Blueprints/test_template.jinja", template_kwargs={"msg":"Hello !"})
    # ---> In Unreal: Will print "Im printed from a custom template ! Hello ! End of template" in the "Output log" tab.
    


import datetime
from dataclasses import dataclass
import json
import logging
import requests
from requests.exceptions import HTTPError
import traceback
import typing
from typing import List, Dict

# Log settings
_LOG_FORMAT = "[%(filename)s:%(lineno)s][%(asctime)s][%(levelname)s] %(message)s"
logging.basicConfig(format=_LOG_FORMAT)
def set_log_level(level):
    assert level in (logging.DEBUG, logging.INFO, logging.ERROR, logging.CRITICAL),\
           f"Invalid log level: {level}"
    logging.getLogger().setLevel(level)

# Default constants
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 30010
DEFAULT_TIMEOUT = 10.0  # seconds
DEFAULT_USE_PROPERTIES_CACHE = False

# URConnection routes
ROUTE_INFOS        = {"method":"get", "route":"remote/info"}
ROUTE_SEARCH_ASSET = {"method":"put", "route":"remote/search/assets"}
ROUTE_BATCH        = {"method":"put", "route":"remote/batch"}
ROUTE_GET_PRESETS  = {"method":"get", "route":"remote/presets"}
ROUTE_GET_PRESET   = {"method":"get", "route":"remote/preset/${PRESET_NAME}"}

# URObject routes
ROUTE_FUNC_CALL = {"method":"put", "route":"remote/object/call"}
ROUTE_PROPERTY  = {"method":"put", "route":"remote/object/property"}
ROUTE_THUMBNAIL = {"method":"put", "route":"remote/object/thumbnail"}
ROUTE_DESCRIBE  = {"method":"put", "route":"remote/object/describe"}

# URemotePreset routes
ROUTE_PRESET_GET_PROPERTY     = {"method":"get", "route":"remote/preset/${PRESET_NAME}/property/${PROPERTY_NAME}"}
ROUTE_PRESET_SET_PROPERTY     = {"method":"put", "route":"remote/preset/${PRESET_NAME}/property/${PROPERTY_NAME}"}
ROUTE_PRESET_RUN_FUNCTION     = {"method":"put", "route":"remote/preset/${PRESET_NAME}/function/${FUNCTION_NAME}"}
ROUTE_PRESET_GET_METADATA     = {"method":"get", "route":"remote/preset/${PRESET_NAME}/metadata"}
ROUTE_PRESET_SET_METADATA     = {"method":"put", "route":"remote/preset/${PRESET_NAME}/metadata/${METADATA_KEY}"}
ROUTE_PRESET_GET_METADATA_KEY = {"method":"get", "route":"remote/preset/${PRESET_NAME}/metadata/${METADATA_KEY}"}
ROUTE_PRESET_DELETE_METADATA  = {"method":"delete", "route":"remote/preset/${PRESET_NAME}/metadata/${METADATA_KEY}"}

# NOT SUPPORTED ATM, WebControl.EnableExperimentalRoutes = 1 needed in DefaultEngine.ini file.
__ROUTE_EVENT = {"method":"put", "route":"remote/object/event"}

# Util paths
EDITOR_ACTOR_SUBSYSTEM = "/Script/UnrealEd.Default__EditorActorSubsystem"
EDITOR_ASSET_LIBRARY = "/Script/EditorScriptingUtilities.Default__EditorAssetLibrary"

def get_version():
    import upyrc
    return upyrc.__version__

# Exceptions
class URConnectionError(Exception): ...
class URConnectionInvalidMethodError(Exception): ...
class URConnectionRouteNotFoundError(Exception): ...
class URConnectionInvalidRequestError(Exception): ...
class UObjectMethodNotFOund(Exception): ...

# ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
# URConnection
# ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
class URConnection:
    ''' Base URConnection object, to prepare an object to send requests to Unreal from.
        This is also the class needed to instanciate _URemoteObject.

        connection = URConnection()
        remote_obj = connection.get_ruobject(obj_path)

        It needs to have the remote API plugin installed on your Unreal project, as well as the server started, you can do that on the cmd:
        WebControl.StartServer
        WebControl.EnableServerOnStartup
    '''
    
    # Ellipsis used to forced a request's result to be ignored, when queued in a batch context object.
    BATCH_CONTEXT_SKIP_RESULT = ...

    def __init__(self, host=DEFAULT_HOST, port=DEFAULT_PORT, timeout=DEFAULT_TIMEOUT, generate_transaction=True):

        self._timeout = timeout
        self._generate_transaction = generate_transaction
        self._host = host
        self._port = port
        self._adress_root = f"http://{self._host}:{self._port}"

        self._batch_context = None

    # --- Core ---

    def run_request(self, route_infos: dict=None, json: dict=None, timeout: float=None):
        ''' Send a request to UE webserver.
        '''
        if timeout is None:
            timeout = self._timeout

        route = route_infos["route"]
        method = route_infos["method"]

        adress = f"{self._adress_root}/{route}"
        logging.debug(f"Running request: {route_infos}")
        logging.debug(f"  json: {json}")
        try:
            method = getattr(requests, method)
        except AttributeError:
            raise URConnectionInvalidMethodError(f"Method {method} not supported.")
        else:
            if json:
                json["generateTransaction"] = self._generate_transaction

            # Has a batch context, delay the request execution, it will be handled by _URConnectionBatchContext
            if isinstance(self._batch_context, _URConnectionBatchContext):
                self._batch_context._add_request(route_infos, json)
                return self.BATCH_CONTEXT_SKIP_RESULT
            else:
                try:
                    result = method(adress, json=json, timeout=timeout, headers={'Content-Type': 'application/json'})
                except requests.exceptions.ConnectionError:
                    logging.error("No connection could be made to: " + adress)
                    raise URConnectionError("No connection could be made to: " + adress)
                return self.handle_request_result(result)

    def handle_request_result(self, result):
        ''' Get the raw request's result, and try to parse it to json or usable data.
            Return and log proper error message if any.
        '''
        logging.debug(f"Request execution time {result.elapsed}.")

        if result.status_code == 404:
            msg = result.json().get("errorMessage", "Unknown error.")
            logging.error(msg)
            raise URConnectionRouteNotFoundError(msg)
        
        if result.status_code == 200:
            if result.content == b'': return None
            try:
                r = result.json()
                r["__request_time_elapsed"] = result.elapsed
                return r
            except json.JSONDecodeError:
                return result.content

        msg = result.json().get("errorMessage", "Unknown error.")
        logging.error(msg)
        raise URConnectionInvalidRequestError(msg)

    def batch_context(self):
        ''' Used to create a batch python context object, where all the call made within that context will be queued.
            Once the context.execute() is called, all the result will be run in a single batch on the web server.

            See _URConnectionBatchContext for more infos.
        '''
        return _URConnectionBatchContext(connection=self)

    def infos(self, timeout=None):
        ''' Return webserver infos, all the calls possible and their parameters. 
        '''
        infos = self.run_request(route_infos=ROUTE_INFOS, timeout=timeout)
        return infos

    def ping(self, timeout: float=5.0) -> datetime.timedelta:
        ''' Simple ping the server, return a timedelta, or None in case of failure. 
        '''
        try:
            return self.infos(timeout=timeout)["__request_time_elapsed"]
        except (URConnectionRouteNotFoundError, URConnectionInvalidRequestError, URConnectionError) as e:
            logging.debug("Ping failed: " + str(e))
            return None

    # --- Getters ---

    def query(self, query: str="", package_names: List[str]=[], class_names: List[str]=[], package_paths: List[str]=[], recursive_classe_exlusion_set: List[str]=[],
                    recursive_path: bool=False, recursive_classes: bool=False):
        ''' Query an Asset in the registry according to filters:

            query: The text you want to match in an asset's name. Leaving this field an empty string returns all results.
            package_names: Restricts the search to exact matches of the specified packages. The format of the value is an array of strings, such as ["/Game/MyFolder/MyAsset"].
            class_names: Restricts the search to exact matches of the specified class names. The format of the value is an array of strings.
            package_paths: Restricts the search to any packages that contain the specified paths. For example, ["/Game/MyFolder"] will only return assets with "/Game/MyFolder" in their asset path. The format of the value is an array of strings.
            recursive_classe_exlusion_set: When searching for matching classes recursively, you can specify any classes you want to exclude from the results. The format of the value is an array of strings.
            recursive_path: A boolean specifying whether to search subfolders of the paths specified in PackagePaths.
            recursive_classes: A boolean specifying whether to look at subclasses of classes specified in ClassNames. 
        '''

        query_filter = {
                "PackageNames":package_names,
                "ClassNames": class_names,
                "PackagePaths": package_paths,
                "RecursiveClassesExclusionSet": recursive_classe_exlusion_set,
                "RecursivePaths": recursive_path,
                "RecursiveClasses": recursive_classes
                }
        body = {"Query":query, "Filter":query_filter}
        result = self.run_request(route_infos=ROUTE_SEARCH_ASSET, json=body)
        if not result: return []
        return result["Assets"]

    def get_ruobject(self, path: str, timeout: float= None, use_properties_cache=DEFAULT_USE_PROPERTIES_CACHE):
        ''' Mandatory function to run to create an _URemoteObject object, from its path.
            See _URemoteObject for more infos.
        '''
        body = {"objectPath": path, "access":"READ_ACCESS"}
        try:
            properties = self.run_request(route_infos=ROUTE_PROPERTY, json=body, timeout=timeout)
        except URConnectionInvalidRequestError as e:
            return None

        try:
            body = {"objectPath": path}
            describe = self.run_request(route_infos=ROUTE_DESCRIBE, json=body, timeout=timeout)
        except URConnectionInvalidRequestError as e:
            return None

        describe["properties"] = properties
        return _URemoteObject(path=path, connection=self, describe=describe, use_properties_cache=use_properties_cache, _allow_create=True)

    def get_preset(self, preset_name: str, timeout: float=None):
        ''' Mandatory function to run to create an _URemotePreset object, from its name.
            See _URemotePreset for more infos.
        '''
        route_infos = ROUTE_GET_PRESET.copy()
        route_infos["route"] = route_infos["route"].replace("${PRESET_NAME}", preset_name)
        try:
            preset_infos = self.run_request(route_infos=route_infos, timeout=timeout)
        except URConnectionRouteNotFoundError as e:
            return None

        preset_infos = preset_infos["Preset"]
        return _URemotePreset(path=preset_infos["Path"], connection=self, describe=preset_infos, use_properties_cache=False, _allow_create=True)

    def get_all_presets(self, timeout: float= None) -> List:
        ''' Get all remote preset name and paths.
        '''
        result = self.run_request(route_infos=ROUTE_GET_PRESETS, timeout=timeout)
        if result:
            return result["Presets"]
        return

    def get_thumbnail(self, asset_path='') -> Dict:
        ''' Get an thumbnail image from an asset path. 
        '''
        assert asset_path != '', "Invalid asset path."
        body = {"objectPath":asset_path}
        try:
            return self.run_request(route_infos=ROUTE_THUMBNAIL, json=body, timeout=self._timeout)
        except URConnectionRouteNotFoundError:
            return None
    
    def get_editor_actor_library(self):
        ''' Get the EditorActorSubsystem, more infos:
            https://docs.unrealengine.com/5.0/en-US/PythonAPI/class/EditorActorSubsystem.html
        '''
        return self.get_ruobject(EDITOR_ACTOR_SUBSYSTEM)

    def get_editor_asset_library(self):
        ''' Get the EditorAssetLibrary, more infos:
            https://docs.unrealengine.com/4.27/en-US/PythonAPI/class/EditorAssetLibrary.html
        '''
        return self.get_ruobject(EDITOR_ASSET_LIBRARY)

@dataclass
class BatchResult:

    request_id: int
    response_code: int
    response_body: dict

    def __str__(self):

        return f"Id: {self.request_id}, {self.response_code} {self.response_body}"

class _URConnectionBatchContext:
    ''' Python context object, used to queue all the requests and execute them with .execute()
        in one batch call.
    '''
    def __init__(self, connection:URConnection = None, timeout:int = 60):
        
        self._connection = connection
        self._context_timeout = timeout
        self._requests = []

    def __enter__(self):
        self._connection._batch_context = self
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        self._connection._batch_context = None

    def _add_request(self, route_infos={}, body={}):
        ''' Called at a request execution, instead of running it, add it to the batch queue.
        '''
        if "generateTransaction" in body:
            del body["generateTransaction"]

        req = {}
        req["RequestId"] = len(self._requests) + 1
        req["URL"] = '/' + route_infos["route"]
        req["Verb"] = route_infos["method"].upper()
        req["Body"] = body

        self._requests.append(req)

    def execute(self) -> List[BatchResult]:
        ''' Execute all requests saved in contect's queue, in one batch request.
            Return an array of BatchResult, with each individual request's result.
        '''
        adress = f"{self._connection._adress_root}/{ROUTE_BATCH['route']}"
        batch_body = {"Requests":self._requests}
        method = getattr(requests, ROUTE_BATCH["method"])

        logging.info(f"Executing {len(self._requests)} requests in batch.")
        
        batch_result = method(adress, json=batch_body, timeout=self._context_timeout,
                              headers={'User-Agent': 'X-UnrealEngine-Agent',
                                       'Content-Type': 'application/json',
                                       'Passphrase': 'dummytext'})
        result = self._connection.handle_request_result(batch_result)
        
        out_data = []
        for response in result["Responses"]:
            out_data.append(BatchResult(response["RequestId"],
                                        response["ResponseCode"],
                                        response["ResponseBody"]))
        return out_data

# ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
# _URemoteObject
# ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

class _URemoteObjectCacheContext:
    ''' Context object to disable / enable _URemoteObject use_properties_cache property. 
    '''
    def __init__(self, uobject=None, use_cache=True):

        self._uobject = uobject
        self._use_cache = use_cache
        self._existing_use_cache = uobject._use_properties_cache

    def __enter__(self):
        self._uobject._use_properties_cache = self._use_cache

    def __exit__(self, ex_type, ex_value, ex_traceback):
        self._uobject._use_properties_cache = self._existing_use_cache

class _URemoteObject:
    ''' Base object to access UObject properties and function remotely.

        -> This class can not be instanciated, the user must use URConnection.get_object(path:str) to get a new instance.

        All _URemoteObject uses an internal cache for properties and functions informations, fetched from the UObject in Unreal.
        All properties can be access normally, as uobejct.my_property (get and set).

        Be default, the properties cache is disabled (see DEFAULT_USE_PROPERTIES_CACHE const). So each time you get a properties or a function info, a new requests will be sent to Unreal.
        If the properties cache is in place, then the property values / functions infos will be fetch from an internal class's cache (similar to __dict__).
        In case of property, if the property is not found in the cache, a requests is made anyway.

        The cache behaviour can be changed in multiple ways, such as:

            - On class instantiation:
                        URConnection.get_object(path, use_properties_cache=True)

            - On class instance:
                        uobject.set_use_properties_cache(True)

            - With a context object, from the class instance:
                        with uobject.properties_cache_context(use_cache=True):
                            print(uobject.my_property)
    '''

    def __new__(cls, *args, **kwargs):
        if not kwargs.get("_allow_create"):
            raise Exception("_URemoteObject can't be instanciated, use URConnection.get_object instead.")
        return object.__new__(cls)

    def __init__(self, path: str='', connection: URConnection=None, describe: dict=None, use_properties_cache=False, **kwargs):
        
        self._use_properties_cache = use_properties_cache
        self._properties_cache = {}

        self._timeout = connection._timeout
        self._connection = connection
        self._name = describe.get("Name")
        self._uclass = describe.get("Class")
        self._functions = self._init_func_dict(describe)
        self._path = path

        self._properties_cache.update(self._init_property_list(describe))

    def __dir__(self) -> List[str]:

        attrs = list(self.__dict__.keys()) + \
                [s + " (function)" for s in list(self._functions.keys())] + \
                [s + " (property)" for s in list(self._properties_cache.keys())]
        return attrs

    def __getattr__(self, name):
        ''' Custom get attribute, for normal properties, the object tries first a requests to the remote object.
            Unless _use_properties_cache is True, then it's fetched from the internal cache directly. 
        '''
        # Handle dunners
        if(name.startswith("__") and name.endswith("__")):
            return super().__getattribute__(name)

        # Check on the cache if needed.
        if super().__getattribute__("_use_properties_cache"):
            try:
                return super().__getattribute__("_properties_cache")[name]
            except KeyError:
                pass
        
        # Run an actual request to get the property from the object in unreal.
        try:
            r = self.get_property(name)
        except URConnectionInvalidRequestError:
            raise AttributeError(name)
        else:
            self._properties_cache[name] = r
            return r

    def __setattr__(self, name, value):
        ''' Custom set attribute, for property not in the current object's built-in dict,
            run a request on the remote object to set the property's value.
        '''
        # If it's an internal attrib, set it normally
        if name in self.__dict__ or name.startswith('_'):
            self.__dict__[name] = value

        # Send a set property request
        else:
            r = self.set_property(property_name=name, property_value=value)
            if r and self._use_properties_cache:
                self._properties_cache[name] = value

    def __str__(self) -> str:

        return f"UObject: {self.get_path()}, of Class {self.get_uclass()}"

    def __eq__(self, other) -> bool:

        try:
            return other.get_path() == self.get_path()
        except AttributeError:
            return False

    def __ne__(self, other) -> bool:

        try:
            return other.get_path() != self.get_path()
        except AttributeError:
            return False

    # --- Core ---

    def _init_property_list(self, describe: dict):
        ''' From a raw describe dictionnary, populate the properties internal cache.
        '''
        return describe["properties"]

    def _init_func_dict(self, describe: dict):
        ''' From a raw describe dictionnary, populate the functions internal cache.
        '''
        funcs = {}
        for f in describe.get("Functions", []):
            
            f_name =  f["Name"]
            if f_name.startswith("K2_"):
                f_name = f_name[3:]
                del f["Name"]
            funcs[f_name] = f

        return funcs

    def properties_cache_context(self, use_cache=True):
        ''' Return a python context object, to allow the user to set the use cache to True or False for all calls in this context.
            For instance:

            with uobject.properties_cache_context(use_cache=True):
                print(uobject.my_property)
        '''
        return _URemoteObjectCacheContext(uobject=self, use_cache=use_cache)

    def refresh_cache(self):
        ''' Run again the describe call, to populate propertiy list. 
        '''
        properties = self.get_all_properties()
        self._properties_cache = properties

    def run_request(self, **kwargs):

        return self._connection.run_request(**kwargs)

    def run_function(self, function_name='', timeout=None, **kwargs):
        ''' Run a function accessible on the remote object, by its name. Any args can be passed as keyword args.
            Function from blueprints are accessible as well.
        '''
        if not function_name in self.get_all_functions():
            UObjectMethodNotFOund(function_name)

        if not timeout:
            timeout = self._timeout
        body = {
                "objectPath":self.get_path(), "access":"WRITE_ACCESS",
                "functionName":function_name,
               }

        if kwargs != {}:
            body["parameters"] = kwargs

        r = self.run_request(route_infos=ROUTE_FUNC_CALL, json=body, timeout=timeout)
        if r:
            return r.get("ReturnValue", r)
        return r

    # --- Getters ---

    def get_name(self) -> str:
        ''' Get name of the current UObject.
        '''
        return self._name

    def get_root_path(self) -> str:
        ''' Get the root path of the UObject, without package.
        '''
        return '/'.join(self._path.split('/')[:-1])

    def get_package(self) -> str:
        ''' Get the package of the current UObject, without path.
        '''
        return self._path.split('/')[-1].split(':')[0].split('.')[0]

    def get_full_name(self) -> str:
        ''' Get UObject full name.
        '''
        return self._path.split(':')[-1]

    def get_uclass(self) -> str:
        ''' Get the class of the UObject (as string).
        '''
        return self._uclass

    def get_path(self) -> str:
        ''' Get the full path of the UObject, including root and package.
        '''
        return self._path

    def get_all_functions(self):
        
        return self._functions

    def get_property(self, property_name='', default=None):
        ''' Get a UObject property value by its name.
            It can be access directly as object's property, like: UObject.my_property
        '''
        if self._use_properties_cache:
            if property_name in self._properties_cache:
                return self._properties_cache[property_name]
            return default

        body = {"objectPath":self.get_path(), "access":"READ_ACCESS", "propertyName": property_name}
        try:
            property_result = self.run_request(route_infos=ROUTE_PROPERTY, json=body, timeout=self._timeout)
            
            # Get property run in a batch, so nothing return here.
            if property_result is URConnection.BATCH_CONTEXT_SKIP_RESULT: return None

            return property_result[property_name]

        except URConnectionInvalidRequestError:
            return default

    def get_function_help(self, function_name: str='', print_out=True):
        ''' Get a function help, name, description and arguments, from its name.
        '''
        func = self.get_all_functions().get(function_name)
        if not func:
            raise UObjectMethodNotFOund(function_name)

        if print_out:
            print('-' * 64)
            print(f"Function: {function_name}")
            print('-' * 64)
            print(func.get("Description", "No description."))
            print("Arguments:")
            for args in func.get("Arguments", []):
                for k, v in args.items():
                    print(f"\t{k}: \t\t{v}")
                print(' ')
            print('-' * 64)

        return func

    def get_all_properties(self):
        ''' Return all object property's names.
        '''
        if self._use_properties_cache:
            if self._properties_cache != {}:
                return self._properties_cache.copy()

        body = {"objectPath":self.get_path(), "access":"READ_ACCESS"}
        all_properties = self.run_request(route_infos=ROUTE_PROPERTY, json=body, timeout=self._timeout)
        if self._use_properties_cache:
            self._properties_cache.update(all_properties)
        return all_properties

    # --- Setters ---

    def set_use_properties_cache(self, use_cache: bool, refresh_cache: bool=True):
        ''' Manually enable / disable cache system on the current object.
        '''
        self._use_properties_cache = use_cache
        if use_cache and (refresh_cache or self._cache_properties == {}):
            self.refresh_cache()
    
    def set_property(self, property_name: str='', property_value=None, timeout: float=None):
        ''' Set a property value from its name and a value.
        '''
        if not timeout:
            timeout = self._timeout
        body = {
                "objectPath":self.get_path(), "access":"WRITE_ACCESS",
                "propertyName":property_name,
                "propertyValue":{property_name:property_value}
               }
        r = self.run_request(route_infos=ROUTE_PROPERTY, json=body, timeout=timeout)
        if r is URConnection.BATCH_CONTEXT_SKIP_RESULT: return None

        if r and r.status_code == 200:
            self._properties_cache[property_name] = property_value
            return property_value
        return False

    # --- Misc ---

    def has_property(self, property_name: str='') -> bool:
        ''' Return True or False if a property with the given name is found on the object.
        '''
        if self._use_properties_cache:
            property_found = property_name in self.__properties_cache
            return property_found
        
        property_found = self.get_property(property_name=property_name)        
        return property_found

    def describe(self):
        ''' Return full description of the remote object, used to populate properties and function lists.
        '''
        body = {"objectPath": self.get_path()}
        return self._connection.run_request(route_infos=ROUTE_DESCRIBE, json=body, timeout=self._timeout)

    def flush_cache(self):
        ''' Flush the internal cache of the RUObject.
        '''
        self._use_properties_cache = False
        self._properties_cache = {}

# ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
# _URemotePreset
# ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

@dataclass(eq=False)
class _URemotePresetProperty:
    ''' Data class to represent a property fetch from a preset. Can be used to set and eval its value. 
    '''
    ID: str
    display_name: str
    underlying_property: dict
    metadata: dict
    owners: list
    preset_name: str
    group: str
    connection: URConnection

    def __eq__(self, other):
        return other.ID == self.ID

    def __str__(self):
        return f"_URemotePresetProperty: {self.display_name} (preset: {self.preset_name})"

    def eval(self):
        ''' Get the property value.
        '''
        route_infos = ROUTE_PRESET_GET_PROPERTY.copy()
        route_infos["route"] = route_infos["route"].replace("${PRESET_NAME}", self.preset_name)\
                                                   .replace("${PROPERTY_NAME}", self.display_name)
        result = self.connection.run_request(route_infos=route_infos)
        if result:
            return result["PropertyValues"][0]["PropertyValue"]
        return None

    def set(self, **kwargs):
        ''' Set the property value.
        '''
        assert kwargs != {}, "Invalid parameters"
        route_infos = ROUTE_PRESET_SET_PROPERTY.copy()
        route_infos["route"] = route_infos["route"].replace("${PRESET_NAME}", self.preset_name)\
                                                   .replace("${PROPERTY_NAME}", self.display_name)
        body = {"PropertyValue": kwargs}
        result = self.connection.run_request(route_infos=route_infos, json=body)

@dataclass(eq=False)
class _URemotePresetFunction:
    ''' Dataclass to represent a function exposed in a preset, can be executed by using the method .run()
    '''
    ID: str
    display_name: str
    underlying_function: dict
    owners: list
    preset_name: str
    group: str
    connection: URConnection

    def __eq__(self, other):
        return other.ID == self.ID

    def __str__(self):
        return f"_URemotePresetFunction: {self.display_name} (preset: {self.preset_name})"

    def run(self, **kwargs):
        
        route_infos = ROUTE_PRESET_RUN_FUNCTION.copy()
        route_infos["route"] = route_infos["route"].replace("${PRESET_NAME}", self.preset_name)\
                                                   .replace("${FUNCTION_NAME}", self.display_name)
        body = {"Parameters":kwargs}
        return self.connection.run_request(route_infos=route_infos, json=body)

@dataclass(eq=False)
class _URemotePresetGroup:
    ''' Dataclass to represent a preset's group, which can contains exposed properties, actors or functions.
    '''
    name: str
    properties: dict
    functions: dict
    actors: dict
    preset_name: str

    def __eq__(self, other):
        return other.name == self.name

    def __str__(self):
        return f"_URemotePresetGroup: {self.name} (preset: {self.preset_name})"

    def get_property(self, property_display_name: str):

        return self.properties.get(property_display_name)

    def get_function(self, function_display_name: str):

        return self.functions.get(function_display_name)

    def get_actor(self, actor_name: str):

        return self.actors.get(actor_name)

@dataclass(eq=False)
class _URemotePresetActor:
    ''' Dataclass to represent an exposed actor in a preset.
    '''
    ID: str
    display_name: str
    underlying_actor: dict
    preset_name: str
    group: str

    def __eq__(self, other):
        return other.ID == self.ID

    def __str__(self):
        return f"_URemotePresetActor: {self.display_name} ({self.actor_path})"

    @property
    def actor_path(self):
        return self.underlying_actor["Path"]

    @property
    def actor_class(self):
        return self.underlying_actor["Class"]

    @property
    def actor_name(self):
        return self.underlying_actor["Name"]

class _URemotePreset(_URemoteObject):
    ''' Represent and Unreal remote preset. You can access properties, functions exposed in the preset, as well as metadata, actors, or groups.
        
        You can get a preset object from a URConnection:
        preset = connection.get_preset(preset_name)

        Or all presets available:
        presets = connection.get_all_presets()

        Unlike the URObject, you can not access properties directly, but with these simple calls:

        preset_property = preset.get_property("property_display_name")
        preset_property.eval()  # To get the value.
        preset_property.set(X=100, Y=10, Z=0)  # To set values.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._uclass = "_URemotePreset"
        self._groups = self._init_group_list(kwargs["describe"])
        self._actors = self._init_actor_list(kwargs["describe"])

    def __getattr__(self, name):
        return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)

    # --- Internal inits ---
        ''' Internal init function, to init properties, functions and exposed actors dict objects.
        '''
    def _init_property(self, group: str, property_infos: dict) -> _URemotePresetProperty:

        prop = _URemotePresetProperty(property_infos["ID"],
                                      property_infos["DisplayName"],
                                      property_infos["UnderlyingProperty"],
                                      property_infos["Metadata"],
                                      property_infos["OwnerObjects"],
                                      self.get_name(),
                                      group, self._connection)
        return prop

    def _init_function(self, group: str, function_infos: dict) -> _URemotePresetFunction:

        func = _URemotePresetFunction(function_infos["ID"],
                                      function_infos["DisplayName"],
                                      function_infos["UnderlyingFunction"],
                                      function_infos["OwnerObjects"],
                                      self.get_name(),
                                      group, self._connection)
        return func

    def _init_group(self, group_infos: dict) -> _URemotePresetGroup:

        grp_name = group_infos["Name"]
        properties = {prop["DisplayName"]:self._init_property(grp_name, prop) for prop in group_infos.get("ExposedProperties", [])}
        functions = {func["DisplayName"]:self._init_function(grp_name, func) for func in group_infos.get("ExposedFunctions", [])}
        actors = {act["DisplayName"]:self._init_actor(grp_name, act) for act in group_infos.get("ExposedActors", [])}
        grp = _URemotePresetGroup(grp_name, properties, functions, actors, self.get_name())
        return grp

    def _init_actor(self, group: str, actor_infos: dict) -> _URemotePresetActor:

        actor = _URemotePresetActor(actor_infos["ID"],
                                    actor_infos["DisplayName"],
                                    actor_infos["UnderlyingActor"],
                                    self.get_name(),
                                    group)
        return actor

    def _init_group_list(self, describe):

        return {grp["Name"]:self._init_group(grp) for grp in describe.get("Groups", [])}

    def _init_actor_list(self, describe):

        actors = {}
        for grp in describe.get("Groups", []):
            for actor in grp.get("ExposedActors", []):
                actors[actor["DisplayName"]] = self._init_actor(grp["Name"], actor)
        return actors

    def _init_property_list(self, describe):

        properties = {}
        for grp in describe.get("Groups", {}):
            for p in grp.get("ExposedProperties", []):
                prop = self._init_property(grp, p)
                properties[prop.display_name] = prop
        return properties

    def _init_func_dict(self, describe: dict):
        
        funcs = {}
        for grp in describe.get("Groups", []):
            for f in grp.get("ExposedFunctions", []):
                func = self._init_function(grp["Name"], f)
                funcs[func.display_name] = func
        return funcs

    # --- Setter ---

    def set_metadata(self, metadata_key: str, metadata_value: str):
        ''' Set metadata value (str) from its key. If it doesn't exist, it will be created.
        '''
        route_infos = ROUTE_PRESET_SET_METADATA.copy()
        route_infos["route"] = route_infos["route"].replace("${PRESET_NAME}", self.get_name())\
                                                   .replace("${METADATA_KEY}", metadata_key)
        body = {"Value":str(metadata_value)}

        return self.run_request(route_infos=route_infos, json=body, timeout=self._timeout)

    # --- Getter ---

    def get_metadata(self, metadata_key: str=None):
        ''' Get metadata from its key (optional), if no key is set, return the whole metadata dictionary.
        '''
        if not metadata_key:
            route_infos = ROUTE_PRESET_GET_METADATA.copy()
            route_infos["route"] = route_infos["route"].replace("${PRESET_NAME}", self.get_name())
        else:
            route_infos = ROUTE_PRESET_GET_METADATA_KEY.copy()
            route_infos["route"] = route_infos["route"].replace("${PRESET_NAME}", self.get_name())\
                                                    .replace("${METADATA_KEY}", metadata_key)

        result = self.run_request(route_infos=route_infos, timeout=self._timeout)
        if result and metadata_key:
            return result["Value"]
        else:
            return result["Metadata"]

    def get_all_groups(self) -> Dict[str, _URemotePresetGroup]:
        ''' Get all groups in the preset, return a list of _URemotePresetGroup.
        '''
        if not self._use_properties_cache:
            self.refresh()
        return self._groups

    def get_group(self, group_name: str) -> _URemotePresetGroup:
        ''' Get a specitic group, according to its name, returns a _URemotePresetGroup.
        '''
        grp = self.get_all_groups().get(group_name)
        return(grp)

    def get_all_actors(self) -> List[_URemotePresetActor]:
        ''' Get all actors exposed to the preset (if any), return a list of _URemotePresetActor.
        '''
        if not self._use_properties_cache:
            self.refresh()
        return self._actors

    def get_actor(self, actor_display_name: str) -> _URemotePresetActor:
        ''' Get a specific exposed actor on this preset, returns a _URemotePresetActor.
        '''
        return self.get_all_actors().get(actor_display_name)

    def get_all_property_names(self) -> List[str]:
        ''' Get all the property names exposed on the preset.
        '''
        if not self._use_properties_cache:
            self.refresh()
        return list(self._properties_cache.keys())

    def get_property(self, property_display_name: str) -> _URemotePresetProperty:
        ''' Get an exposed property on the preset, retuns a _URemotePresetProperty.
            From that object, value can be fetch using .eval(), or set, using .set(value).
        '''
        if not self._use_properties_cache:
            self.refresh()
        return self._properties_cache.get(property_display_name)

    def get_all_function_name(self) -> List[str]:
        ''' Get all exposed function's display name.
        '''
        if not self._use_properties_cache:
            self.refresh()
        return list(self._functions.keys())

    def get_function(self, function_name: str) -> _URemotePresetFunction:
        ''' Get a function exposed on the preset, by display name.
        '''
        if not self._use_properties_cache:
            self.refresh()
        return self._functions.get(function_name)

    # --- Misc ---

    def remove_metadata(self, metadata_key: str):
        ''' Remove a metadata entry from the preset, from its key.
        '''
        route_infos = ROUTE_PRESET_DELETE_METADATA.copy()
        route_infos["route"] = route_infos["route"].replace("${PRESET_NAME}", self.get_name())\
                                                   .replace("${METADATA_KEY}", metadata_key)

        return self.run_request(route_infos=route_infos, timeout=self._timeout)

    def has_group(self, group_name: str) -> bool:
        ''' Return True if the given group_name is found in the preset's groups.
        '''
        return group_name in list(self.get_groups().keys())

    def refresh(self, timeout: float=DEFAULT_TIMEOUT) -> None:
        '''Run a describe call on the preset again. To repopulate properties and functions as well as groups and actors.
        '''
        route_infos = ROUTE_GET_PRESET.copy()
        route_infos["route"] = route_infos["route"].replace("${PRESET_NAME}", self.get_name())
        describe = self.run_request(route_infos=route_infos, timeout=timeout)["Preset"]
        self._functions = self._init_func_dict(describe)
        self._properties_cache = self._init_property_list(describe)
        self._groups = self._init_group_list(describe)
        self._actors = self._init_actor_list(describe)




import os
import configparser
import uuid
import socket
import tempfile
import logging
import json
from jinja2 import Environment, PackageLoader, FileSystemLoader
from pathlib import Path
from typing import Union, List

from upyrc.upyre_json_pipe import upyre_json_pipe

'''

This module allows a user to execute python code remotely on a unreal engine session.
It needs to have the python plugin enabled on unreal, as well as the allow python remote execution option set on.

In order to create a connection, we need to fetch the settings of the python plugin, this can be done by giving the full path to a uproject to the create config object:

config = upyre.RemoteExecutionConfig.from_uproject_path(r"D:\Projects\SandBox53\SandBox53.uproject")

You can also create the config manually with the right values, such as multicast_group ip and port, multicast_bind_adress or ip_muticast_ttl.
Those values can be found in the project settings of the python plugin.
config = upyre.RemoteExecutionConfig(mutlicast_group=("239.0.0.1", 6766), multicast_bind_adress="0.0.0.0")

To open a connection, you have to use PythonRemoteConnection object, it can be used as context manager like this:

with upyre.PythonRemoteConnection(config) as conn:
    result = conn.execute_python_command("unreal.log('hello form python !')", exec_type=upyre.ExecTypes.EXECUTE_FILE, raise_exc=True)
    print(result)

Or you can call open_connection() and close_connection() manually when needed.

To check help on exec_types, you can call help(ExecTypes). Help is available on other objects as well.

A json base communication pipeline is available, it can be enabled with: upyre.PythonRemoteConnection(config, open_json_output_pipe=True).
This will create a temporary json file where data can be written to on Unreal session side and are automatically available after command execution on the PythonCommandResult object.
Data can be written from the PythonRemoteConnection as well using the write(entry_name, entry_data) method, it can be then read by Unreal session.
See upyre_json_pipe/upyre_json_pipe.py for more infos.

'''

def get_version():
    import upyrc
    return upyrc.__version__

# Log settings
_LOG_FORMAT = "[%(filename)s:%(lineno)s][%(asctime)s][%(levelname)s] %(message)s"
logging.basicConfig(format=_LOG_FORMAT)
def set_log_level(level):
    assert level in (logging.DEBUG, logging.INFO, logging.ERROR, logging.CRITICAL),\
           f"Invalid log level: {level}"
    logging.getLogger().setLevel(level)

# Constants
UE_MAGIC = "ue_py"
PROTOCOL_VERSION = 1
SOCKET_TIMEOUT = 0.5 # second
PYTHON_SETTING_INI_FILE = "DefaultEngine.ini"
PYTHON_SETTING_ENTRY = "/Script/PythonScriptPlugin.PythonScriptPluginSettings"  # read only in DefaultEngine.ini in projects
CONNECTION_UUID_SIZE = 4

# Jinja templates env
jinja_env = Environment(loader=PackageLoader("upyrc", "re_templates"))

# Command output pipe default
COMMAND_JSON_OUTPUT_PIPE_DEFAULT_FILE = tempfile.gettempdir() + '\\upyrc_output_pipe.json'
AUTO_REMOVE_JSON_TEMP_FILE = True
JSON_OUTPUT_PIPE_MODULE_FOLDER = os.path.dirname(__file__) + "\\upyre_json_pipe"

# Command ip and port to send commands between python and unreal.
def get_command_address():
    with socket.socket() as s:
        s.bind(('', 0))
        _, port = s.getsockname()
    return ("127.0.0.1", port)

# Config fallbacks from env
UPYRE_BUFFER_SIZE = os.environ.get("UPYRE_BUFFER_SIZE", "2_097_152")
UPYRE_BUFFER_SIZE = int(UPYRE_BUFFER_SIZE)
UPYRE_MULTICAST_GROUP_ADDR = os.environ.get("UPYRE_MULTICAST_GROUP_ADDR", "239.0.0.1")
UPYRE_MULTICAST_GROUP_PORT = os.environ.get("UPYRE_MULTICAST_GROUP_PORT", "6766")
UPYRE_MULTICAST_GROUP_PORT = int(UPYRE_MULTICAST_GROUP_PORT)
UPYRE_MULTICAST_BIND_ADDRESS = os.environ.get("MULTICAST_BIND_ADDRESS", "127.0.0.1")
UPYRE_IP_MULTICAST_TTL = os.environ.get("UPYRE_IP_MULTICAST_TTL", "0")
UPYRE_IP_MULTICAST_TTL = int(UPYRE_IP_MULTICAST_TTL)

# For config fallbacks
BUFFER_SIZE = UPYRE_BUFFER_SIZE
MULTICAST_GROUP = (UPYRE_MULTICAST_GROUP_ADDR, UPYRE_MULTICAST_GROUP_PORT)
MULTICAST_BIND_ADDRESS = UPYRE_MULTICAST_BIND_ADDRESS
IP_MULTICAST_TTL = UPYRE_IP_MULTICAST_TTL

# Python exec types enum
class ExecTypes:
    ''' "ExecuteFile" => Execute the Python command as a file. This allows you to execute either a literal Python script containing multiple statements, or a file with optional arguments.
        "ExecuteStatement" => Execute the Python command as a single statement. This will execute a single statement and print the result. This mode cannot run files.
        "EvaluateStatement" => Evaluate the Python command as a single statement. This will evaluate a single statement and return the result. This mode cannot run files.
    '''
    EXECUTE_FILE = "ExecuteFile"
    EXECUTE_STATEMENT = "ExecuteStatement"
    EVALUATE_STATEMENT = "EvaluateStatement"

# Config exceptions
class InvalidUprojectPathError(Exception): ...
class InvalidConfigError(Exception): ...

# Connection exception
class ConnectionError(Exception): ...
class OutputPipeNotOpenedError(Exception): ...

class RemoteExecutionConfig:
    ''' Config used to create connection to unreal. There are 2 ways to create it:

        config = RemoteExecutionConfig(buffer_size=0, multicast_group=(), project_name='TestProject' ...) the user can set it the config manually, if values are not set, this will use the fallback.
        config = RemoteExecutionConfig.from_uproject_path('project_path.uproject'), this will parse the settings found in the project directly.
                                                                                   A full path to a uproject must be given (str).
    '''
    def __init__(self, buffer_size=BUFFER_SIZE, ip_multicast_ttl=IP_MULTICAST_TTL,
                       multicast_group=MULTICAST_GROUP, multicast_bind_address=MULTICAST_BIND_ADDRESS,
                       project_name='', local_uuid=''):

        self.BUFFER_SIZE = buffer_size
        self.MULTICAST_GROUP = multicast_group
        self.MULTICAST_BIND_ADDRESS = multicast_bind_address
        self.IP_MULTICAST_TTL = ip_multicast_ttl

        # Unique indentifier for local process
        if not local_uuid:
            local_uuid = str(uuid.uuid4())
        self.LOCAL_UUID = local_uuid

        # Command ip and port to send commands between python and unreal.
        self.COMMAND_ADDRESS = get_command_address()

        # Optional, project name (without .uproject) will be used to match the right unreal instance if multiple are opened.
        # If not set, the first instance will be used.
        self.PROJECT_NAME = project_name

    def __str__(self) -> str:
        
        return f"Config values: BUFFER_SIZE: {self.BUFFER_SIZE}, MULTICAST_GROUP: {self.MULTICAST_GROUP}, MULTICAST_BIND_ADDRESS: {self.MULTICAST_BIND_ADDRESS}, IP_MULTICAST_TTL: {self.IP_MULTICAST_TTL}, COMMAND_ADDRESS: {self.COMMAND_ADDRESS}, PROJECT_NAME: {self.PROJECT_NAME}"

    @classmethod
    def from_uproject_path(cls, uproject_path: Path):
        ''' Config will be created from 'DefaultEngine.ini' found from the given .uproject path.
            Config's 'project_name' will be automatically set as well.
            If the Python remote execution is not enable, it will raise a InvalidConfigError.
        '''
        uproject_path = Path(uproject_path)
        if not uproject_path.is_file() or not uproject_path.suffix == ".uproject":
            raise InvalidUprojectPathError(str(uproject_path))
        
        unreal_config = uproject_path.parent / "Config" / PYTHON_SETTING_INI_FILE
        if not unreal_config.exists():
            raise InvalidUprojectPathError(f"Can't find: {unreal_config}")
        
        parser = configparser.ConfigParser(strict=False)
        parser.read(unreal_config)

        settings = parser[PYTHON_SETTING_ENTRY]
        remote_execution_enabled = settings.get("bRemoteExecution")
        if not remote_execution_enabled:
            raise InvalidConfigError("Python remote execution not enable in python project settings.")
        
        multicast_group = settings.get("RemoteExecutionMulticastGroupEndpoint")
        if not multicast_group:
            multicast_group = MULTICAST_GROUP
        else:
            multicast_group = (multicast_group.split(':')[0], int(multicast_group.split(':')[1]))

        multicast_bind_address = settings.get("RemoteExecutionMulticastBindAddress", MULTICAST_BIND_ADDRESS)
        ip_multicast_ttl = int(settings.get("RemoteExecutionMulticastTtl", IP_MULTICAST_TTL))
        buffer_size = int(settings.get("RemoteExecutionReceiveBufferSizeBytes", BUFFER_SIZE))

        project_name = uproject_path.name.split('.')[0]

        config = RemoteExecutionConfig(buffer_size=buffer_size, multicast_group=multicast_group,
                                       ip_multicast_ttl=ip_multicast_ttl, multicast_bind_address=multicast_bind_address,
                                       project_name=project_name)
        return config

class _Message:
    ''' Base classe for all message sent to unreal
    '''
    TYPE = ''

    def __init__(self, config: RemoteExecutionConfig):
        self._raw_data = {}
        self.config = config

    def to_data(self):
        return json.dumps(self._raw_data).encode()
    
    def send(self, s: socket.socket):
        ''' Send encoded json data to unreal node.
        '''
        logging.debug(f"Sending message of type '{self.TYPE}'")
        data_msg = self.to_data()
        s.sendto(data_msg, self.config.MULTICAST_GROUP)

    def receive(self, s: socket.socket):
        ''' How the data is reveived needs to be implemented in each message type.
        '''
        raise NotImplementedError()

    def raw_receive(self, s: socket.socket):
        ''' Receive all data from unreal, this might have multiple result (for ping calls for instance).
            It returns an iterator, an it's up to each message type to handle the result properly.
        '''
        data_received = b''
        while 1:
            try:
                data, _ = s.recvfrom(BUFFER_SIZE)
                data_received += data

                try:
                    json_data = json.loads(data_received)
                    data_received = b''
                except json.decoder.JSONDecodeError:
                    continue
                if json_data["type"] == self.TYPE:
                    continue # ignore echo.

                yield json_data

            except socket.timeout:
                break

class PingMessage(_Message):
    ''' Message sent to unreal to find unreal node(s).
    '''
    TYPE = "ping"

    def __init__(self, config: RemoteExecutionConfig):
        super().__init__(config=config)
        self._raw_data = {
                            "version":PROTOCOL_VERSION,
                            "magic":UE_MAGIC,
                            "source":config.LOCAL_UUID,
                            "type":self.TYPE
                         }
        
    def receive(self, s: socket.socket) -> dict:

        for result in self.raw_receive(s):

            # if config has a project name set, we need to be sure to return the right unreal instance with the right project name.
            if self.config.PROJECT_NAME:
                if result["data"]["project_name"] != self.config.PROJECT_NAME:
                    continue
                return result
            
            # if no project name is set, simply send the first result.
            else:
                return result

class OpenConnectionMessage(_Message):
    ''' Message sent to unreal to open a connection to execute command.
        This will create another socket to send command to unreal and receive output.
    '''
    TYPE = "open_connection"

    def __init__(self, unreal_node_id: str, config: RemoteExecutionConfig):
        super().__init__(config=config)
        self._raw_data = {
                            "type":self.TYPE,
                            "version":PROTOCOL_VERSION,
                            "magic":UE_MAGIC,
                            "source":self.config.LOCAL_UUID,
                            "dest":unreal_node_id,
                            "data":{
                                "command_ip":self.config.COMMAND_ADDRESS[0],
                                "command_port":self.config.COMMAND_ADDRESS[1]
                            }
                         }

class CloseConnectionMessage(_Message):
    ''' Close the connection to unreal commands, it needs the unreal node id.
    '''
    TYPE = "close_connection"
    def __init__(self, unreal_node_id: str, config: RemoteExecutionConfig):
        super().__init__(config=config)
        self._raw_data = {
                            "type":self.TYPE,
                            "version":PROTOCOL_VERSION,
                            "magic":UE_MAGIC,
                            "dest":unreal_node_id,
                            "source":self.config.LOCAL_UUID
                         }

class PythonCommandError(Exception): ...

class PythonCommandResult:

    def __init__(self, json_result_data: dict, raise_exc: bool=False):
        ''' json_result_data: the raw json data received from unreal.
            raise_exc: if True and an exception appened during the command execution, it will raise an PythonCommandError exception.
        '''
        self.data = json_result_data.get("data", {})
        self.source_node_id = json_result_data["source"]
        self.target_node_id = json_result_data["dest"]
        self.raise_exc = raise_exc

        self.output_pipe_data_opened = False
        self.output_pipe_data_file = ''
        self._output_pipe_data = {}

    def __bool__(self) -> bool:
        return self.success
    
    def _retrieve_output_pipe_data(self, temp_file: str=''):
        ''' If output_pipe_data was opened on the PythonRemoteConnection creation, a temporary json file
            is created in order to transfert some data between the two (or more) processes.
            At the end of the script execution, those data are read out of the json file and will be available using PythonCommandResult.output_pipe_data.get("name").

            By default, the json file is then deleted, this behavior can be changed in the module global library AUTO_REMOVE_TEMP_FILE.
        '''
        if not temp_file:
            raise OutputPipeNotOpenedError()
        
        if os.path.exists(temp_file):
            with open(temp_file, 'r') as f:
                self._output_pipe_data = json.load(f)

    @property
    def success(self) -> bool:
        ''' Return if the remote script succeeded or not.
        '''
        success = self.data.get("success", False)
        if not success and self.raise_exc:
            raise PythonCommandError(self.result)
        return success
    
    @property
    def result(self) -> str:
        ''' Returns the string result of the remote script execution.
        '''
        return self.data.get("result", "None")

    @property
    def output(self) -> list:
        ''' Returns the whole standard output of the python script execution.
        '''
        return [o for o in self.data.get("output", [])]
    
    @property
    def output_pipe_data(self):
        return self._output_pipe_data
    
    def has_output_pipe_data(self, data_name: str='') -> bool:
        ''' Return True or False if the output pipe data has an entry defined by data_name (str).
            If no data_name is provided, check if there are any data at all.
        '''
        if not data_name:
            return self._output_pipe_data != {}
        else:
            return data_name in self._output_pipe_data.keys()

    def get_first_output_with_identifier(self, identifier: str) -> str:
        ''' Extracts an string output out of the stdout data, starting with an identifier of format XXX=data.
            In the remote script, you would print out this data simply using print("identifier=data here").
            This method is deprecated, use output_pipe_data instead.
        '''
        if not identifier.startswith("="): identifier = identifier + "="
        for data in self.output:
            if data.get("output", '').startswith(identifier):
                return data.get("output").replace(identifier, '').strip()
        return None

    @property
    def output_str(self) -> str:
        return '\n'.join([o['type'] + ': ' + o['output'] for o in self.output])
    
    def __str__(self):
        if not self.success:
            return self.result
        return self.output_str

class PythonRemoteCommandConnection(_Message):
    ''' Create a connection between unreal and python to execute command on. This needs the unreal node ID fetched by ping message.
    '''
    TYPE = "command"

    def __init__(self, unreal_node_id: str, config: RemoteExecutionConfig):
        super().__init__(config=config)
        self.unreal_node_id = unreal_node_id

        # Create command socket to send and receive data when launching command
        # This is a different socket than the multicast one used to ping and open/close the connection.
        self.cmd_sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
        self.cmd_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.cmd_sock.bind(self.config.COMMAND_ADDRESS)
        self.cmd_sock.settimeout(2.0)
        self.cmd_sock.listen()

        # Connection create with unreal once the message is sent.
        self.cmd_connection, _ = self.cmd_sock.accept()
        self.cmd_connection.settimeout(SOCKET_TIMEOUT)

        self._raw_data = {
                          "type":self.TYPE,
                          "version":PROTOCOL_VERSION,
                          "magic":UE_MAGIC,
                          "source":self.config.LOCAL_UUID,
                          "dest":self.unreal_node_id,
                          "data":{}
                         }
        
    def send(self, command: str, exec_type: ExecTypes=ExecTypes.EVALUATE_STATEMENT, unattended: bool=True, timeout: float=5.0, raise_exc: bool=False) -> PythonCommandResult:
        ''' Send the command, the receive method is executed as well to return the result of the command.
        '''
        logging.debug("Sending command.")
        self._raw_data["data"] = {
                                "command":command,
                                "unattended":unattended,
                                "exec_mode":exec_type
        }
            
        data_msg = self.to_data()
        self.cmd_connection.sendto(data_msg, self.config.COMMAND_ADDRESS)

        return self.receive(timeout=timeout, raise_exc=raise_exc)

    def receive(self, timeout: float=5.0, raise_exc: bool=False) -> PythonCommandResult:
        ''' Get the json result data for the executed command, and construct a CommandResult object with it.
            The implementation is a bit different than other messages as the timeout mechanism can't be used here as we don't know how long the command will take unreal side.
            For safety, we have a different timeout version than other message if there is an issue with getting the command output. It can be set on the execute_python_command function.
        '''

        self.cmd_connection.settimeout(timeout)
        json_data = None
        data_received = b''
        while 1:
            try:
                data, _ = self.cmd_connection.recvfrom(BUFFER_SIZE)
                data_received += data

                try:
                    json_data = json.loads(data_received)
                    data_received = b''
                except json.decoder.JSONDecodeError:
                    continue
                if json_data["type"] == self.TYPE:
                    json_data = None
                    continue # ignore echo.

                # Data is complete and valid at this point
                break

            except socket.timeout:
                break

        logging.debug("Command result received.")
        return PythonCommandResult(json_data, raise_exc=raise_exc)

class PythonRemoteConnection:
    ''' Open a connection to a unreal node (by project name, or by default, the first one found, depending on the config).
        From there you can execute a python command.
        It can be used as context manager, to close the connection once out.
    '''
    def __init__(self, config: RemoteExecutionConfig=None, project_name: str='', open_json_output_pipe: bool=False,
                       json_output_pipe_temp_file: str=COMMAND_JSON_OUTPUT_PIPE_DEFAULT_FILE):

        if not config:
            config = RemoteExecutionConfig()
        self.config = config
        logging.debug(str(config))

        if project_name:
            self.config.PROJECT_NAME = project_name

        # Connection uuid, mostly used by json pipe.
        self.__uuid = os.urandom(CONNECTION_UUID_SIZE).hex()

        # Open a pipe where data can be written in a temp json file to be retrieved after command execution.
        self.open_json_output_pipe = open_json_output_pipe
        self.json_output_pipe_temp_file = self.build_json_pipe_file_path(json_output_pipe_temp_file)
        self.json_pipe = None

        self.connection_created = False
        self.unreal_node_id = ''
        self.remote_command_connection = None
        self.conneciton_infos = None

        # Init the multicast socket, to send messages to unreal (Ping, Open/Close connection).
        self.mcastsock = socket.socket(socket.AF_INET,socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self.mcastsock.settimeout(SOCKET_TIMEOUT)
        self.mcastsock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, IP_MULTICAST_TTL)
        self.mcastsock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, 1)
        self.mcastsock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.mcastsock.bind((self.config.MULTICAST_BIND_ADDRESS, self.config.MULTICAST_GROUP[1]))
        membership_request = socket.inet_aton(self.config.MULTICAST_GROUP[0]) + socket.inet_aton(self.config.MULTICAST_BIND_ADDRESS)
        self.mcastsock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, membership_request)

    def build_json_pipe_file_path(self, json_file_path: str) -> str:
        ''' Add connection uuid to json file temp path.
        '''
        r, f = os.path.split(json_file_path)
        return os.path.join(r, self.__uuid + '_' + f)
    
    def get_json_pipe_file_path(self) -> str:
        ''' Get the temp pipe json file path, including the connection's uuid.
        '''
        return self.json_output_pipe_temp_file

    def init_json_pipe(self, json_output_pipe_temp_file: str='', force: bool=False) -> bool:
        ''' Execute the add_command_json_output_pipe.jinja template, which will append the upyre_json_pipe module to unreal python's sys.paths.
        '''

        if self.open_json_output_pipe and not force:
            logging.debug(rf"Json output pipe module already available.")
            return

        self.open_json_output_pipe = True
        if json_output_pipe_temp_file != '':
            json_output_pipe_temp_file = self.build_json_pipe_file_path(json_output_pipe_temp_file)
        if json_output_pipe_temp_file and json_output_pipe_temp_file != self.json_output_pipe_temp_file:
            self.json_output_pipe_temp_file = json_output_pipe_temp_file

        # Add module to sys.paths and update env variables.\
        result = self.execute_template_command("add_command_json_output_pipe.jinja",
                                               template_kwargs={"json_output_pipe_module":JSON_OUTPUT_PIPE_MODULE_FOLDER,
                                                                "json_output_pipe_file":self.json_output_pipe_temp_file})
        if result.success:
            self.json_pipe = upyre_json_pipe.CommandOutputPipe(self.json_output_pipe_temp_file)
            logging.info(rf"Json output pipe module available: {self.json_output_pipe_temp_file}.")
        else:
            self.json_pipe = None
            self.open_json_output_pipe = False
            logging.error(rf"Json output pipe module not available.")

        return result.success
    
    def write(self, entry_name, entry_value):
        ''' Write data to the temp json file pipe.
        '''
        if self.json_pipe is None or not self.open_json_output_pipe:
            raise OutputPipeNotOpenedError()
        self.json_pipe.write(entry_name, entry_value)

    def read(self, entry_name, default=None):
        ''' Read data from the temp json file pipe.
        '''
        if self.json_pipe is None or not self.open_json_output_pipe:
            raise OutputPipeNotOpenedError()
        return self.json_pipe.read(entry_name, default=default)

    def flush(self):
        ''' Flush the pipe data, if any.
        '''
        if self.json_pipe:
            self.json_pipe.flush()

    def __enter__(self):
        self.open_connection()
        if self.open_json_output_pipe:
            self.init_json_pipe(force=True)
        return self

    @property
    def socket(self):
        return self.mcastsock
    
    def __exit__(self, *args):
        self.close_connection()
    
    def open_connection(self):
        ''' Ping unreal until it finds a node (unreal editor instance running) which matches the target project name, if any, otherwise take the first node found.
            When a node is found, open a connection with it by sending an open_connection message and open a receiver socket here too.
        '''
        ping_message = PingMessage(self.config)
        ping_message.send(self.mcastsock)
        pong_data = ping_message.receive(self.mcastsock)
        if not pong_data:
            raise ConnectionError("Connection failed.")
        self.unreal_node_id = pong_data["source"]
        self.connection_infos = pong_data["data"]

        OpenConnectionMessage(self.unreal_node_id, self.config).send(self.mcastsock)
        self.connection_created = True
        self.remote_command_connection = PythonRemoteCommandConnection(self.unreal_node_id, self.config)
        logging.info(f"Connection established: project {pong_data['data']['project_name']} (unreal {pong_data['data']['engine_version']})")

    def close_connection(self):
        ''' Close the active command connection to unreal node.
        '''
        if self.connection_created or self.unreal_node_id != '':
            CloseConnectionMessage(self.unreal_node_id, self.config).send(self.mcastsock)
            self.connection_infos = None
            self.connection_created = False
            self.unreal_node_id = ''
            self.mcastsock.close()
            logging.debug("Connection closed !")

        # Remove objet and temp file involved in json pipe.
        if AUTO_REMOVE_JSON_TEMP_FILE:
            self.json_pipe = None
            try:
                os.remove(self.json_output_pipe_temp_file)
            except OSError:
                pass

    def execute_python_command(self, command: str, exec_type: ExecTypes=ExecTypes.EVALUATE_STATEMENT, unattended: bool=True, timeout: float=5.0, raise_exc: bool=False) -> PythonCommandResult:
        ''' Execute the given python command, this can be either python statement(s) or a path to a python file.
            unattended: Whether to run the command in "unattended" mode (suppressing some UI)
            exec_type: See ExecTypes help.
            timeout: in seconds, defautl is 5. After than timeout, the function result will not be fetched, but it might still be executed on unreal side !
            raise_exc: if True and an exception appened during the command execution, it will raise an PythonCommandError exception.
        '''
        if not self.remote_command_connection:
            raise ConnectionError("Connection was not openned.")
        
        result = self.remote_command_connection.send(command=command, exec_type=exec_type, unattended=unattended, timeout=timeout, raise_exc=raise_exc)

        if self.open_json_output_pipe:
            result._retrieve_output_pipe_data(self.json_output_pipe_temp_file)
            result.output_pipe_data_opened = self.open_json_output_pipe
            result.output_pipe_data_file = self.json_output_pipe_temp_file

        return result
    
    def execute_template_file(self, file_path: Union[Path, str], template_kwargs: dict={},
                              search_paths: List[Union[Path, str]]=[],
                              timeout: float=5.0, raise_exc=True) -> PythonCommandResult:
        ''' Render and execute a given .jinja template file.
            An optionnal list of search paths can be given, to find other inherited templates.
        '''
        
        if isinstance(file_path, str):
            file_path = Path(file_path)

        parent = file_path.parent
        search_paths.append(parent)
        temp_env = Environment(loader=FileSystemLoader([str(s) for s in search_paths]))
        template = temp_env.get_template(file_path.name)
        python_code = template.render(template_kwargs)

        return self.execute_python_command(python_code, exec_type=ExecTypes.EXECUTE_FILE, timeout=timeout, raise_exc=raise_exc)
    
    def execute_template_command(self, template_name: str, timeout: float=5.0, raise_exc: bool=False, exec_type=ExecTypes.EXECUTE_FILE, template_kwargs={}) -> PythonCommandResult:
        ''' Execute a rendered jinja template file, found in re_templates folder.
            'template_kwargs' dicitonnary will be applied to the template.
        '''
        if not template_name.endswith(".jinja"):
            template_name = template_name.split('.')[0] + ".jinja"

        template = jinja_env.get_template(template_name)
        python_code = template.render(template_kwargs)
        return self.execute_python_command(python_code, exec_type=exec_type, timeout=timeout, raise_exc=raise_exc)

    # Helper functions
    
    def log(self, message :str="", raise_exc=True) -> PythonCommandResult:
        return self.execute_template_command("log.jinja", template_kwargs={"msg":message, "level":"message"}, raise_exc=raise_exc)
    
    def log_warning(self, message :str="", raise_exc=True) -> PythonCommandResult:
        return self.execute_template_command("log.jinja", template_kwargs={"msg":message, "level":"warning"}, raise_exc=raise_exc)
    
    def log_error(self, message :str="", raise_exc=True) -> PythonCommandResult:
        return self.execute_template_command("log.jinja", template_kwargs={"msg":message, "level":"error"}, raise_exc=raise_exc)
    
    def execute_bp_method(self, bp_class_path: str="", bp_method_name: str="", args=(), kwargs={}, raise_exc=True) -> PythonCommandResult:
        ''' Execute the template to execute utility BP functions.
        '''
        assert bp_class_path != '', "Invalid bp_class_path."
        assert bp_method_name != '', "Invalid bp_method_name."
        assert bp_class_path.endswith('_C'), "Invalid bp_class_path, must end with '_C'"
        return self.execute_template_command("run_editor_bp_method.jinja",
                                            template_kwargs={"bp_class_path":bp_class_path, "bp_method_name":bp_method_name, "call_method":True, "args":args, "kwargs":kwargs},
                                            raise_exc=raise_exc)
    
    def execute_editor_widget_bp_method(self, bp_path: str="", bp_method_name: str="", args=(), kwargs={}, raise_exc=True) -> PythonCommandResult:
        ''' Execute the template to execute utility BP functions.
        '''
        assert bp_path != '', "Invalid bp_class_path."
        assert bp_method_name != '', "Invalid bp_method_name."
        return self.execute_template_command("run_editor_widget_bp_method.jinja",
                                            template_kwargs={"bp_path":bp_path, "bp_method_name":bp_method_name, "args":args, "kwargs":kwargs},
                                            raise_exc=raise_exc)
    
    def set_bp_property(self, bp_class_path: str="", properties={}, raise_exc=True) -> PythonCommandResult:
        ''' Execute the template to set properties on a utility BP.
        '''
        assert bp_class_path != '', "Invalid bp_class_path."
        assert properties != {}, "Invalid properties."
        assert bp_class_path.endswith('_C'), "Invalid bp_class_path, must end with '_C'"
        return self.execute_template_command("run_editor_bp_method.jinja",
                                             template_kwargs={"bp_class_path":bp_class_path, "properties":properties, "set_properties":True},
                                             raise_exc=raise_exc)

    def spawn_utility_widget_bp(self, widget_path: str="", raise_exc: bool=True) -> PythonCommandResult:
        ''' Take a utility widget BP and spawn it as tab.
        '''
        self.init_json_pipe()
        return self.execute_template_command("spawn_utility_widget_bp.jinja",
                                             template_kwargs={"widget_path":widget_path},
                                             raise_exc=raise_exc)
    
    def close_utility_widget_bp_from_id(self, widget_id: str="", raise_exc: bool=True) -> PythonCommandResult:
        ''' Close a utility widget already spawned.
        '''
        self.init_json_pipe()
        return self.execute_template_command("close_widget_bp_from_id.jinja",
                                             template_kwargs={"widget_id":widget_id},
                                             raise_exc=raise_exc)
    
    def import_fbx_static_meshes(self, fbx_file_paths: List[str], destination_folder_path: str,
                                       combine_meshes: bool=True, raise_exc: bool=True) -> PythonCommandResult:
        ''' Import given fbx file(s) and import them (static meshes only) to the given destination_folder_path.
            The destination_folder_path must starts with /Game/ (Which is the /Content/ of a UE project).

            If multiple meshes are found in the FBX, they can be combined or not, using the combine_meshes bool.
        '''
        self.init_json_pipe()
        return self.execute_template_command("fbx_static_mesh_import.jinja",
                                             template_kwargs={"fbx_file_paths":[f.replace('\\', '/') for f in fbx_file_paths],
                                                              "destination_folder_path":destination_folder_path,
                                                              "combine_meshes":combine_meshes},
                                             raise_exc=raise_exc)
        
    def export_fbx_static_meshes(self, static_mesh_package_names: List[str], destination_folder_path: str,
                                       export_collisions: bool=False, export_level_of_detail: bool=False, export_vertex_color: bool=False,
                                       raise_exc: bool=True) -> PythonCommandResult:
        ''' Export given static mesh(es) from package name (/Game/myfolder/myasset) as fbx in a given folder.
        '''
        self.init_json_pipe()
        return self.execute_template_command("fbx_static_mesh_export.jinja",
                                             template_kwargs={"static_mesh_package_names":static_mesh_package_names,
                                                              "export_collisions":export_collisions,
                                                              "export_level_of_detail":export_level_of_detail,
                                                              "export_vertex_color":export_vertex_color,
                                                              "destination_folder_path":destination_folder_path},
                                             raise_exc=raise_exc)
    


from upyrc import upyrc
#import logging
#upyrc.set_log_level(logging.DEBUG)
print("Version:", upyrc.get_version())

print("\n----------- INIT CONNECTION --------------------------------------------------------------------------------------------------------------------------------\n")

# Create a connection to Unreal
conn = upyrc.URConnection()
print("Ping: ", conn.ping())
# >>> Ping: 0:00:00.015211

print("\n----------- REMOTE OBJECT --------------------------------------------------------------------------------------------------------------------------------\n")

# Access to object using paths
blue_cube_path = "/Game/Levels/L_upyrc.L_upyrc:PersistentLevel.StaticMeshActor_2"

blue_cube = conn.get_ruobject(blue_cube_path)
print("RUObject: " + str(blue_cube))
# >>> RUObject: UObject: /Game/Maps/TestMap.TestMap:PersistentLevel.StaticMeshActor_2, of Class /Script/Engine.StaticMeshActor

# Run a function remotely
folder_path = blue_cube.run_function("GetFolderPath")
print("Folder path: ", folder_path)
# >>> Folder path: TestData

# Access property directly on URemoteObject object:
auto_lod_generation = blue_cube.bEnableAutoLODGeneration
print("bEnableAutoLODGeneration: ", auto_lod_generation)
# >>> bEnableAutoLODGeneration: True

# Access property by name:
auto_lod_generation = blue_cube.get_property("bEnableAutoLODGeneration")
print("bEnableAutoLODGeneration (from func): ", auto_lod_generation)
# >>> bEnableAutoLODGeneration (from func): True

# Set property directly on URemoteObject object:
blue_cube.bEnableAutoLODGeneration = False
print("After set bEnableAutoLODGeneration to False: ", blue_cube.bEnableAutoLODGeneration)
# >>> After set bEnableAutoLODGeneration to False: False

# Set property by name:
blue_cube.set_property("bEnableAutoLODGeneration", True)
print("After set bEnableAutoLODGeneration to True (from func): ", blue_cube.bEnableAutoLODGeneration)
# >>> After set bEnableAutoLODGeneration to True (from func): True

# dir() returns all properties and functions available remotly.
print(dir(blue_cube))
# >>> ['ActorHasTag (function)', 'AddActorLocalOffset (function)', 'AddActorLocalRotation (function)', 'AddActorLocalTransform (function)',  ... 'bRelevantForLevelBounds (property)', 'bRelevantForNetworkReplays (property)', 'bStaticMeshReplicateMovement (property)']

print("\n----------- REMOTE BLUEPRINT --------------------------------------------------------------------------------------------------------------------------------\n")

# You can also load a blueprint actor.
blueprint_path = "/Game/Levels/L_upyrc.L_upyrc:PersistentLevel.BPU_TestActorUtility_C_1"
blueprint = conn.get_ruobject(blueprint_path)
print("Blueprint: ", blueprint)
# >>> UObject: /Game/Maps/TestMap.TestMap:PersistentLevel.BUA_TestBlueprint_C_1, of Class /Game/Blueprints/BUA_TestBlueprint.BUA_TestBlueprint_C

# You can run an editor function from the blueprint.
blueprint.run_function("SimpleTestFunc")
# ---> In Unreal: LogBlueprintUserMessages: [BUA_TestBlueprint_C_UAID_D89EF37396CEAA5E01_2120751343] Hello from BP utility actor ! Let move those tree cubes !
# ---> in Unreal  the three test cubes, SM_BlueCube, SM_RedCube and SM_GreenCube have now a random offset in Z.

# And a function with argument(s) too and a return value:
function_result = blueprint.run_function("SimpleTestFuncArgs", MyString="Hello from python !")
print("Function result: ", function_result)
# ---> In Unreal: LogBlueprintUserMessages: [BUA_TestBlueprint_C_UAID_D89EF37396CEAA5E01_2120751343] Hello from BP utility actor, msg: Hello from python !
# >>> Function result:  {'MessagePrinted': 'Hello from BP utility actor, msg: Hello from python !', '__request_time_elapsed': datetime.timedelta(microseconds=7084)}

print("\n----------- REMOTE PRESET --------------------------------------------------------------------------------------------------------------------------------\n")

# Get all presets basic infos ( Name, ID and path )
all_presets = conn.get_all_presets()
print("Presets: ", all_presets)
# >>> Presets: [{'Name': 'RCP_TestPreset', 'ID': '916A553549DCE66727D00585DED11589', 'Path': '/Game/RCP_TestPreset.RCP_TestPreset'}]

# Get an preset object, by name.
preset_name = "RCP_PresetA"
preset = conn.get_preset(preset_name)
print("Preset: ", preset)
# >>> Preset:  UObject: /Game/RCP_TestPreset.RCP_TestPreset, of Class RemotePreset

# Get a preset exposed property:
preset_property = preset.get_property("Relative Location")
print("Relative Location: ", preset_property.eval())
# >>> Relative Location (SM_Lamp_Ceiling) value: {'X': 330, 'Y': 10, 'Z': 0}

# Set a property value:
preset_property.set(Z=200.0)
# The object "PointLight" has now its Z position set to 200.0

# Changing the PointLight color to red.
preset_property_light_color = preset.get_property("Light Color")
preset_property_light_color.set(B=10,G=10,R=255,A=255)


# Get a function
preset_function = preset.get_function("Function from Preset")
print("Preset function: ", preset_function)
# >>> Preset function:  _URemotePresetFunction: My Lib Func (BPF_testFuncLib_C) (preset: RCP_TestPreset)
preset_function.run()
# ---> In unreal "Hello from lib !"

# Get an exposed actor info, from preset.
preset_exposed_actor = preset.get_actor("SM_BlueCube")
print("Exposed actor: ", preset_exposed_actor)
# >>> Exposed actor:  _URemotePresetActor: SM_TableRound (/Game/Maps/TestMap.TestMap:PersistentLevel.StaticMeshActor_6)

print("\n----------- QUERIES --------------------------------------------------------------------------------------------------------------------------------\n")

# Queries the assets registry. Here, get all objects from it. It doesn't return an UObject array, but and array of directionnaries with all the infos found.
all_objects = conn.query()
print("Query all objects: ", all_objects)
# >>> Query all objects: [{'Name': 'BUA_TestBlueprint_C_UAID_D89EF37396CEAA5E01_2120751343', 'Class': '/Game/Blueprints/BUA_TestBlueprint.BUA_TestBlueprint_C', 'Path': '/Game/Maps/MainMap.MainMap:PersistentLevel.BUA_TestBlueprint_C_UAID_D89EF37396CEAA5E01_2120751343', 'Metadata': {'ActorLabel': 'BUA_TestBlueprint'}}, ...]

# You can add filters, here, query only the staticmeshes:
all_static_meshes = conn.query(class_names=["/Script/Engine.StaticMeshActor"])
print("Query all static meshes: ", all_static_meshes)
# >>> Query all static meshes: Query all static meshes:  [{'Name': 'StaticMeshActor_UAID_A4AE111137DC54FB00_1240666663', 'Class': '/Script/Engine.StaticMeshActor', 'Path': '/Game/Maps/MainMap.MainMap:PersistentLevel.StaticMeshActor_UAID_A4AE111137DC54FB00_1240666663', 'Metadata': {'ActorLabel': 'SM_SkySphere'}}, ...]

# To have more infos on the filters that you can apply, check function's help:
help(conn.query)

print("\n----------- UTILITIES --------------------------------------------------------------------------------------------------------------------------------\n")

# Get special utilities, for instance here, actor library, you can check, using dir() what functions are available.
# You can get all the level actor's paths:
actors_library = conn.get_editor_actor_library()
all_level_actors = actors_library.run_function("GetAllLevelActors")
print("All level actors: ", all_level_actors)
# >>> All level actors: ['/Game/Maps/TestMap.TestMap:PersistentLevel.Floor', '/Game/Maps/TestMap.TestMap:PersistentLevel.DirectionalLight_0', ... ]

# You can also get the asset_library utility. As all other UObjects, you can get a function's help:
asset_library = conn.get_editor_asset_library()
asset_library.get_function_help("DeleteDirectory")


print("\n----------- BATCHING PROCESSES --------------------------------------------------------------------------------------------------------------------------------\n")

#/!\ NOT SUPPORTED AT THE MOMENT, UE5 CRASH ON REMOTE BATCH CALLS /!\

blue_cube_path = "/Game/Levels/L_upyrc.L_upyrc:PersistentLevel.StaticMeshActor_2"
green_cube_path = "/Game/Levels/L_upyrc.L_upyrc:PersistentLevel.StaticMeshActor_3"
red_cube_path = "/Game/Levels/L_upyrc.L_upyrc:PersistentLevel.StaticMeshActor_4"

blue_cube = conn.get_ruobject(blue_cube_path)
green_cube = conn.get_ruobject(green_cube_path)
red_cube = conn.get_ruobject(red_cube_path)

with conn.batch_context() as ctx:

    blue_cube.bEnableAutoLODGeneration = False
    green_cube.bEnableAutoLODGeneration = False
    red_cube.bEnableAutoLODGeneration = False

    results = ctx.execute()

for r in results:
    print(r)

